"""Defines functions common across diffract challenges.

Copyright (c) 2025 invrs.io LLC
"""

import dataclasses
from typing import Any, Optional, Tuple, Union

import fmmax
import jax
import jax.numpy as jnp
import numpy as onp
from jax import tree_util
from totypes import json_utils, types

from invrs_gym import utils
from invrs_gym.challenges import base

Params = Any

DENSITY_LOWER_BOUND = 0.0
DENSITY_UPPER_BOUND = 1.0

THICKNESS_CAP = "thickness_cap"
THICKNESS_GRATING = "thickness_grating"
THICKNESS_SPACER = "thickness_spacer"
DENSITY = "density"

EFIELD = "efield"
HFIELD = "hfield"
FIELD_COORDINATES = "field_coordinates"


@dataclasses.dataclass
class GratingSpec:
    """Defines the physical specifcation of a grating.

    Attributes:
        permittivity_ambient: Permittivity of the ambient material.
        permittivity_cap: Permittivity of the material between grating and ambient.
        permittivity_grating: Permittivity of the grating teeth.
        permittivity_encapsulation: Permittivity of the material in gaps between
            grating teeth.
        permittivity_spacer: Permittivity of the spacer layer between grating and
            substrate.
        permittivity_substrate: Permittivity of the substrate.
        thickness_cap: Thickness of the cap layer between grating and ambient.
        thickness_grating: Thickness of the grating layer.
        thickness_spacer: Thickness of the spacer layer between grating and substrate.
        period_x: The size of the unit cell along the x direction.
        period_y: The size of the unit cell along the y direction.
        grid_spacing: The spacing of the grid on which grating permittivity is defined.
    """

    permittivity_ambient: complex
    permittivity_cap: complex
    permittivity_grating: complex
    permittivity_encapsulation: complex
    permittivity_spacer: complex
    permittivity_substrate: complex

    thickness_ambient: float
    thickness_cap: float | jnp.ndarray | types.BoundedArray
    thickness_grating: float | jnp.ndarray | types.BoundedArray
    thickness_spacer: float | jnp.ndarray | types.BoundedArray
    thickness_substrate: float

    period_x: float
    period_y: float

    grid_spacing: float

    @property
    def grid_shape(self) -> Tuple[int, int]:
        """Return the shape of the grid implied by `grid_spacing`."""
        return (
            int(jnp.ceil(self.period_x / self.grid_spacing)),
            int(jnp.ceil(self.period_y / self.grid_spacing)),
        )


@dataclasses.dataclass
class GratingSimParams:
    """Parameters that configure the simulation of a grating.

    Attributes:
        wavelength: The wavelength of the excitation.
        polar_angle: The polar angle of the excitation.
        azimuthal_angle: The azimuthal angle of the excitation.
        formulation: The FMM formulation to be used.
        approximate_num_terms: Defines the number of terms in the Fourier expansion.
        truncation: Determines how the Fourier basis is truncated.
    """

    wavelength: float | jnp.ndarray
    polar_angle: float | jnp.ndarray
    azimuthal_angle: float | jnp.ndarray
    formulation: fmmax.Formulation
    approximate_num_terms: int
    truncation: fmmax.Truncation


@dataclasses.dataclass
class GratingResponse:
    """Contains the response of the grating.

    Attributes:
        wavelength: The wavelength for the efficiency calculation.
        polar_angle: The polar angle for the efficiency calculation.
        azimuthal_angle: The azimuthal angle for the efficiency calculation.
        transmission_efficiency: The per-order and per-wavelength coupling efficiency
            with which the excitation is transmitted.
        transmission_efficiency: The per-order and per-wavelength coupling efficiency
            with which the excitation is reflected.
        expansion: Defines the Fourier expansion for the calculation.
    """

    wavelength: jnp.ndarray
    polar_angle: jnp.ndarray
    azimuthal_angle: jnp.ndarray
    transmission_efficiency: jnp.ndarray
    reflection_efficiency: jnp.ndarray
    expansion: fmmax.Expansion


json_utils.register_custom_type(GratingResponse)

tree_util.register_pytree_node(
    GratingResponse,
    lambda r: (
        (
            r.wavelength,
            r.polar_angle,
            r.azimuthal_angle,
            r.transmission_efficiency,
            r.reflection_efficiency,
            r.expansion,
        ),
        None,
    ),
    lambda _, children: GratingResponse(*children),
)


# -----------------------------------------------------------------------------
# Define components used in diffract optimization challenges.
# -----------------------------------------------------------------------------


class SimpleGratingComponent(base.Component):
    """A simple grating component whose only optimizable parameter is density."""

    def __init__(
        self,
        spec: GratingSpec,
        sim_params: GratingSimParams,
        density_initializer: base.DensityInitializer,
        **seed_density_kwargs: Any,
    ) -> None:
        """Initializes the grating component.

        Args:
            spec: Defines the physical specification of the grating.
            sim_params: Defines simulation parameters for the grating.
            density_initializer: Callable which generates the initial density from
                a random key and the seed density.
            **seed_density_kwargs: Keyword arguments which set the attributes of
                the seed density used to generate the inital parameters.
        """

        self.spec = spec
        self.sim_params = sim_params
        self.seed_density = seed_density(
            grid_shape=self.spec.grid_shape,
            **seed_density_kwargs,
        )
        self.density_initializer = density_initializer

        self.expansion = fmmax.generate_expansion(
            primitive_lattice_vectors=fmmax.LatticeVectors(
                u=self.spec.period_x * fmmax.X,
                v=self.spec.period_y * fmmax.Y,
            ),
            approximate_num_terms=self.sim_params.approximate_num_terms,
            truncation=self.sim_params.truncation,
        )

    def init(self, key: jax.Array) -> types.Density2DArray:
        """Return the initial parameters for the grating component."""
        params = self.density_initializer(key, self.seed_density)
        # Ensure that there are no weak types in the initial parameters.
        return tree_util.tree_map(
            lambda x: jnp.asarray(x, jnp.asarray(x).dtype), params
        )

    def response(
        self,
        params: types.Density2DArray,
        *,
        wavelength: Optional[Union[float, jnp.ndarray]] = None,
        expansion: Optional[fmmax.Expansion] = None,
        compute_fields: bool = False,
    ) -> Tuple[GratingResponse, base.AuxDict]:
        """Computes the response of the grating.

        The response consists of the transmitted and reflected power into each order
        for both TE- and TM-polarized plane wave illumination.

        Args:
            params: The parameters defining the metagrating, matching those returned
                by the `init` method.
            wavelength: Optional wavelength to override the default in `sim_params`.
            expansion: Optional expansion to override the default `expansion`.
            compute_fields: If `True`, computes and xz cross section for electric
                and magnetic fields, which makes the calculation more expensive.

        Returns:
            The `(response, aux)` tuple.
        """
        if expansion is None:
            expansion = self.expansion
        if wavelength is None:
            wavelength = self.sim_params.wavelength
        (transmission_efficiency, reflection_efficiency), aux = grating_efficiency(
            density=params,
            spec=self.spec,
            wavelength=jnp.asarray(wavelength),
            polar_angle=jnp.asarray(self.sim_params.polar_angle),
            azimuthal_angle=jnp.asarray(self.sim_params.azimuthal_angle),
            expansion=expansion,
            formulation=self.sim_params.formulation,
            compute_fields=compute_fields,
        )
        response = GratingResponse(
            wavelength=jnp.asarray(wavelength),
            polar_angle=jnp.asarray(self.sim_params.polar_angle),
            azimuthal_angle=jnp.asarray(self.sim_params.azimuthal_angle),
            transmission_efficiency=transmission_efficiency,
            reflection_efficiency=reflection_efficiency,
            expansion=expansion,
        )
        return response, aux


class GratingWithOptimizableThicknessComponent(base.Component):
    """A grating whose optimizable parameters are density and layer thickness."""

    def __init__(
        self,
        spec: GratingSpec,
        sim_params: GratingSimParams,
        thickness_initializer: base.ThicknessInitializer,
        density_initializer: base.DensityInitializer,
        **seed_density_kwargs: Any,
    ) -> None:
        """Initializes the grating component splitter component.

        Args:
            spec: Defines the physical specification of the grating.
            sim_params: Defines simulation parameters for the grating.
            thickness_initializer: Callable which returns the initial thickness for
                the grating layer from a random key and a bounded array with value
                equal the thickness from `spec`.
            density_initializer: Callable which generates the initial density from
                a random key and the seed density.
            **seed_density_kwargs: Keyword arguments which set the attributes of
                the seed density used to generate the inital parameters.
        """

        self.spec = spec
        self.sim_params = sim_params
        self.thickness_initializer = thickness_initializer
        self.density_initializer = density_initializer

        self.seed_density = seed_density(
            grid_shape=self.spec.grid_shape,
            **seed_density_kwargs,
        )

        self.expansion = fmmax.generate_expansion(
            primitive_lattice_vectors=fmmax.LatticeVectors(
                u=self.spec.period_x * fmmax.X,
                v=self.spec.period_y * fmmax.Y,
            ),
            approximate_num_terms=self.sim_params.approximate_num_terms,
            truncation=self.sim_params.truncation,
        )

    def init(self, key: jax.Array) -> Params:
        """Return the initial parameters for the grating component."""
        keys = iter(jax.random.split(key, num=4))
        params = {
            THICKNESS_CAP: self.thickness_initializer(
                next(keys), self.spec.thickness_cap  # type: ignore[arg-type]
            ),
            THICKNESS_GRATING: self.thickness_initializer(
                next(keys), self.spec.thickness_grating  # type: ignore[arg-type]
            ),
            THICKNESS_SPACER: self.thickness_initializer(
                next(keys), self.spec.thickness_spacer  # type: ignore[arg-type]
            ),
            DENSITY: self.density_initializer(next(keys), self.seed_density),
        }
        # Ensure that there are no weak types in the initial parameters.
        return tree_util.tree_map(
            lambda x: jnp.asarray(x, jnp.asarray(x).dtype), params
        )

    def response(
        self,
        params: Params,
        *,
        wavelength: Optional[Union[float, jnp.ndarray]] = None,
        expansion: Optional[fmmax.Expansion] = None,
        compute_fields: bool = False,
    ) -> Tuple[GratingResponse, base.AuxDict]:
        """Computes the response of the grating component.

        The response consists of the transmitted and reflected power into each order
        for both TE- and TM-polarized plane wave illumination.

        Args:
            params: The parameters defining the grating, matching those returned by
                the `init` method.
            wavelength: Optional wavelength to override the default in `sim_params`.
            expansion: Optional expansion to override the default `expansion`.
            compute_fields: If `True`, computes and xz cross section for electric
                and magnetic fields, which makes the calculation more expensive.

        Returns:
            The `(response, aux)` tuple.
        """
        if expansion is None:
            expansion = self.expansion
        if wavelength is None:
            wavelength = self.sim_params.wavelength
        spec = dataclasses.replace(
            self.spec,
            thickness_cap=jnp.asarray(params[THICKNESS_CAP].array),
            thickness_grating=jnp.asarray(params[THICKNESS_GRATING].array),
            thickness_spacer=jnp.asarray(params[THICKNESS_SPACER].array),
        )
        (transmission_efficiency, reflection_efficiency), aux = grating_efficiency(
            density=params[DENSITY],  # type: ignore[arg-type]
            spec=spec,
            wavelength=jnp.asarray(wavelength),
            polar_angle=jnp.asarray(self.sim_params.polar_angle),
            azimuthal_angle=jnp.asarray(self.sim_params.azimuthal_angle),
            expansion=expansion,
            formulation=self.sim_params.formulation,
            compute_fields=compute_fields,
        )
        response = GratingResponse(
            wavelength=jnp.asarray(wavelength),
            polar_angle=jnp.asarray(self.sim_params.polar_angle),
            azimuthal_angle=jnp.asarray(self.sim_params.azimuthal_angle),
            transmission_efficiency=transmission_efficiency,
            reflection_efficiency=reflection_efficiency,
            expansion=expansion,
        )
        return response, aux


def seed_density(grid_shape: Tuple[int, int], **kwargs: Any) -> types.Density2DArray:
    """Return the seed density for a grating component.

    Args:
        grid_shape: The shape of the grid on which the density is defined.
        kwargs: keyword arguments specifying additional properties of the seed
            density, e.g. symmetries.

    Returns:
        The seed density.
    """

    # Check kwargs that are required for a grating component.
    invalid_kwargs = ("array", "lower_bound", "upper_bound", "periodic")
    if any(k in invalid_kwargs for k in kwargs):
        raise ValueError(
            f"Attributes were specified which confict with automatically-extracted "
            f"attributes. Got {kwargs.keys()} when {invalid_kwargs} are automatically "
            f"extracted."
        )

    mid_density_value = (DENSITY_LOWER_BOUND + DENSITY_UPPER_BOUND) / 2
    return types.Density2DArray(
        array=jnp.full(grid_shape, mid_density_value),
        lower_bound=DENSITY_LOWER_BOUND,
        upper_bound=DENSITY_UPPER_BOUND,
        periodic=(True, True),
        **kwargs,
    )


def index_for_order(
    order: Tuple[int, int],
    expansion: fmmax.Expansion,
) -> int:
    """Returns the index for the specified Fourier order and expansion."""
    ((order_idx,),) = onp.where(onp.all(expansion.basis_coefficients == order, axis=1))
    assert tuple(expansion.basis_coefficients[order_idx, :]) == order
    return int(order_idx)


# -----------------------------------------------------------------------------
# Simulation method used by all gratings.
# -----------------------------------------------------------------------------


def grating_efficiency(
    density: types.Density2DArray,
    spec: GratingSpec,
    wavelength: jnp.ndarray,
    polar_angle: jnp.ndarray,
    azimuthal_angle: jnp.ndarray,
    expansion: fmmax.Expansion,
    formulation: fmmax.Formulation,
    compute_fields: bool,
) -> Tuple[Tuple[jnp.ndarray, jnp.ndarray], base.AuxDict]:
    """Compute the per-order transmission and reflection efficiency for a grating.

    The excitation for the calculation are separate a TE- and TM-polarized plane waves
    at the specified wavelength(s), incident from the substrate with the specified
    polar and azimuthal angles.

    Args:
        density: Defines the pattern of the grating layer.
        spec: Defines the physical specifcation of the grating.
        wavelength: The wavelength of the excitation.
        polar_angle: The polar angle of the excitation.
        azimuthal_angle: The azimuthal angle of the excitation.
        expansion: Defines the Fourier expansion for the calculation.
        formulation: Defines the FMM formulation to be used.
        compute_fields: If `True`, computes fields in an xz cross section.

    Returns:
        The per-order transmission and reflection efficiency, having shape
        `(num_wavelengths, num_fourier_terms, 2)`.
    """
    density_array = utils.transforms.rescaled_density_array(
        density,
        lower_bound=DENSITY_LOWER_BOUND,
        upper_bound=DENSITY_UPPER_BOUND,
    )
    permittivities = (
        jnp.full((1, 1), spec.permittivity_ambient),
        jnp.full((1, 1), spec.permittivity_cap),
        utils.transforms.interpolate_permittivity(
            permittivity_solid=jnp.asarray(spec.permittivity_grating),
            permittivity_void=jnp.asarray(spec.permittivity_encapsulation),
            density=density_array,
        ),
        jnp.full((1, 1), spec.permittivity_spacer),
        jnp.full((1, 1), spec.permittivity_substrate),
    )

    in_plane_wavevector = fmmax.plane_wave_in_plane_wavevector(
        wavelength=wavelength,
        polar_angle=polar_angle,
        azimuthal_angle=azimuthal_angle,
        # Polar angle is defined in substrate, since light is incident from substrate.
        permittivity=jnp.asarray(spec.permittivity_substrate),
    )
    layer_solve_results = [
        fmmax.eigensolve_isotropic_media(
            wavelength=jnp.asarray(wavelength),
            in_plane_wavevector=in_plane_wavevector,
            primitive_lattice_vectors=fmmax.LatticeVectors(
                u=spec.period_x * fmmax.X,
                v=spec.period_y * fmmax.Y,
            ),
            permittivity=p,
            expansion=expansion,
            formulation=formulation,
        )
        for p in permittivities
    ]

    # Layer thicknesses for the ambient and substrate are set to zero; these do not
    # affect the result of the calculation.
    layer_thicknesses = (
        jnp.asarray(spec.thickness_ambient),
        jnp.asarray(spec.thickness_cap),
        jnp.asarray(spec.thickness_grating),
        jnp.asarray(spec.thickness_spacer),
        jnp.asarray(spec.thickness_substrate),
    )

    if compute_fields:
        # If fields wanted, compute the full set of interior scattering matrices.
        s_matrices_interior = fmmax.stack_s_matrices_interior(
            layer_solve_results=layer_solve_results,
            layer_thicknesses=layer_thicknesses,
        )
        s_matrix = s_matrices_interior[-1][0]
    else:
        s_matrix = fmmax.stack_s_matrix(
            layer_solve_results=layer_solve_results,
            layer_thicknesses=layer_thicknesses,
        )

    n = expansion.num_terms
    assert tuple(expansion.basis_coefficients[0, :]) == (0, 0)
    assert expansion.basis_coefficients.shape[0] == n

    # Generate the wave amplitudes for backward-going TE or TM-polarized plane waves
    # at the end of substrate layer.
    bwd_amplitude_substrate_end = jnp.zeros((2 * n, 2), dtype=complex)
    bwd_amplitude_substrate_end = bwd_amplitude_substrate_end.at[0, 0].set(1.0)  # TE
    bwd_amplitude_substrate_end = bwd_amplitude_substrate_end.at[n, 1].set(1.0)  # TM

    # Calculate the incident power from the substrate. Since the substrate thickness
    # has been set to zero, the forward and backward amplitudes are already colocated.
    fwd_amplitude_substrate_start = s_matrix.s12 @ bwd_amplitude_substrate_end
    fwd_flux_substrate, bwd_flux_substrate = fmmax.amplitude_poynting_flux(
        forward_amplitude=fwd_amplitude_substrate_start,
        backward_amplitude=bwd_amplitude_substrate_end,
        layer_solve_result=layer_solve_results[-1],
    )

    # Sum over orders and polarizations to get the total incident flux.
    total_incident_flux = jnp.sum(bwd_flux_substrate, axis=-2, keepdims=True)

    # Calculate the transmitted power in the ambient.
    bwd_amplitude_ambient_end = s_matrix.s22 @ bwd_amplitude_substrate_end
    _, bwd_flux_ambient = fmmax.amplitude_poynting_flux(
        forward_amplitude=jnp.zeros_like(bwd_amplitude_ambient_end),
        backward_amplitude=bwd_amplitude_ambient_end,
        layer_solve_result=layer_solve_results[0],
    )

    # Sum the fluxes over the two polarizations for each order. Note that the sum is
    # done for each excitation condition (i.e. TE and TM), so as to capture the
    # effect of polarization conversion.
    bwd_flux_ambient = bwd_flux_ambient[..., :n, :] + bwd_flux_ambient[..., n:, :]
    fwd_flux_substrate = fwd_flux_substrate[..., :n, :] + fwd_flux_substrate[..., n:, :]

    transmission_efficiency = bwd_flux_ambient / total_incident_flux
    reflection_efficiency = fwd_flux_substrate / total_incident_flux

    # -------------------------------------------------------------------------
    # Compute fields in an xz cross section.
    # -------------------------------------------------------------------------

    aux = {}
    if compute_fields:
        amplitudes_interior = fmmax.stack_amplitudes_interior(
            s_matrices_interior=s_matrices_interior,
            forward_amplitude_0_start=jnp.zeros_like(bwd_amplitude_substrate_end),
            backward_amplitude_N_end=bwd_amplitude_substrate_end,
        )
        x = jnp.linspace(0, spec.period_x, density.shape[0])
        y = jnp.ones_like(x) * spec.period_y / 2
        layer_znum = tuple(
            [int(jnp.round(t / spec.grid_spacing) + 1) for t in layer_thicknesses]
        )
        (ex, ey, ez), (hx, hy, hz), (x, y, z) = fmmax.stack_fields_3d_on_coordinates(
            amplitudes_interior=amplitudes_interior,
            layer_solve_results=layer_solve_results,
            layer_thicknesses=layer_thicknesses,
            layer_znum=layer_znum,
            x=x,
            y=y,
        )
        aux.update(
            {
                EFIELD: (ex, ey, ez),
                HFIELD: (hx, hy, hz),
                FIELD_COORDINATES: (x, y, z),
            }
        )

    return (transmission_efficiency, reflection_efficiency), aux



"""Defines the bayer color sorter component.

Copyright (c) 2025 invrs.io LLC
"""

import dataclasses
import functools
from typing import Any, Callable, Dict, Optional, Tuple

import fmmax
import jax
import jax.numpy as jnp
import numpy as onp
from jax import tree_util
from totypes import json_utils, types

from invrs_gym import utils
from invrs_gym.challenges import base
from invrs_gym.utils import materials

NDArray = onp.ndarray[Any, Any]


Params = Dict[str, types.BoundedArray | types.Density2DArray]
ThicknessInitializer = Callable[[jax.Array, types.BoundedArray], types.BoundedArray]


THICKNESS_METASURFACE = "thickness_metasurface"
DENSITY_METASURFACE = "density_metasurface"
OFFSET_MONITOR_SUBSTRATE = "offset_monitor_substrate"

EFIELD_XY = "efield_xy"
HFIELD_XY = "hfield_xy"
POYNTING_FLUX_XY = "poynting_flux_xy"
COORDINATES_XY = "coordinates_xy"
TRANSMITTED_POWER = "transmitted_power"

EFIELD_XZ = "efield_xz"
HFIELD_XZ = "hfield_xz"
POYNTING_FLUX_XZ = "poynting_flux_xz"
COORDINATES_XZ = "coordinates_xz"

DENSITY_LOWER_BOUND = 0.0
DENSITY_UPPER_BOUND = 1.0


@dataclasses.dataclass
class BayerSpec:
    """Defines the physical specification of a bayer color sorter.

    Attributes:
        material_ambient: The ambient material.
        material_metasurface_solid: The material comprising metasurface solid regions.
        material_metasurface_void: The material comprising metasurface void regions.
        material_substrate: The substrate material.
        thickness_ambient: The thickness of the ambient.
        thickness_metasurface: Thicknesses of the metasurface layers.
        thickness_substrate: The thickness of the substrate.
        offset_monitor_substrate: Offset of the monitor plane from the interface
            between spacer and substrate.
        pixel_size: The size of a pixel in the bayer array.
        grid_spacing: The spacing of the grid on which grating permittivity is defined.
    """

    material_ambient: materials.Material
    material_metasurface_solid: materials.Material
    material_metasurface_void: materials.Material
    material_substrate: materials.Material

    thickness_ambient: float
    thickness_metasurface: types.BoundedArray
    thickness_substrate: float
    offset_monitor_substrate: types.BoundedArray

    pixel_size: float
    grid_spacing: float

    @property
    def period_x(self) -> float:
        """Returns the size of the unit cell in the `x` direction."""
        return 2 * self.pixel_size

    @property
    def period_y(self) -> float:
        """Returns the size of the unit cell in the `y` direction."""
        return 2 * self.pixel_size

    @property
    def grid_shape(self) -> Tuple[int, int]:
        """Return the shape of the grid implied by `grid_spacing`."""
        return (
            int(jnp.ceil(self.period_x / self.grid_spacing)),
            int(jnp.ceil(self.period_y / self.grid_spacing)),
        )


@dataclasses.dataclass
class BayerSimParams:
    """Parameters that configure the simulation of a bayer color sorter.

    Attributes:
        wavelength: The wavelength of the excitation.
        polar_angle: The polar angle of the excitation.
        azimuthal_angle: The azimuthal angle of the excitation.
        formulation: The FMM formulation to be used.
        approximate_num_terms: Defines the number of terms in the Fourier expansion.
        truncation: Determines how the Fourier basis is truncated.
    """

    wavelength: float | jnp.ndarray
    polar_angle: float | jnp.ndarray
    azimuthal_angle: float | jnp.ndarray
    formulation: fmmax.Formulation
    approximate_num_terms: int
    truncation: fmmax.Truncation


@dataclasses.dataclass
class BayerResponse:
    """Contains the response of the bayer color sorter.

    Attributes:
        wavelength: The wavelength for the sorter response.
        polar_angle: The polar angle for the sorter response.
        azimuthal_angle: The azimuthal angle for the sorter response.
        transmission: The transmission into the four pixels for each polarization.
        reflection: The reflection back to the ambient for the each polarizations.
    """

    wavelength: jnp.ndarray
    polar_angle: jnp.ndarray
    azimuthal_angle: jnp.ndarray
    transmission: jnp.ndarray
    reflection: jnp.ndarray


json_utils.register_custom_type(BayerResponse)

tree_util.register_pytree_node(
    BayerResponse,
    lambda r: (
        (
            r.wavelength,
            r.polar_angle,
            r.azimuthal_angle,
            r.transmission,
            r.reflection,
        ),
        None,
    ),
    lambda _, children: BayerResponse(*children),
)


class BayerComponent(base.Component):
    """Defines a bayer color sorter component."""

    def __init__(
        self,
        spec: BayerSpec,
        sim_params: BayerSimParams,
        thickness_initializer: ThicknessInitializer,
        density_initializer: base.DensityInitializer,
        **seed_density_kwargs: Any,
    ) -> None:
        """Initializes the bayer color sorter component.

        Args:
            spec: Defines the physical specification of the bayer color sorter.
            sim_params: Defines simulation parameters for the sorter.
            thickness_initializer: Callable which returns the initial thickness for
                a layer from a random key and a bounded array with value equal the
                thickness from `spec`.
            density_initializer: Callable which generates the initial density from
                a random key and the seed density.
            **seed_density_kwargs: Keyword arguments which set the attributes of
                the seed density used to generate the inital parameters.
        """

        self.spec = spec
        self.sim_params = sim_params
        self.thickness_initializer = thickness_initializer
        self.density_initializer = density_initializer

        self.seed_density = _seed_density(
            grid_shape=self.spec.grid_shape, **seed_density_kwargs
        )
        self.expansion = fmmax.generate_expansion(
            primitive_lattice_vectors=fmmax.LatticeVectors(
                u=self.spec.period_x * fmmax.X,
                v=self.spec.period_y * fmmax.Y,
            ),
            approximate_num_terms=self.sim_params.approximate_num_terms,
            truncation=self.sim_params.truncation,
        )

    def init(self, key: jax.Array) -> Params:
        """Return the initial parameters for the bayer color sorter component."""
        key_thickness, key_density, key_offset = jax.random.split(key, num=3)
        params = {
            THICKNESS_METASURFACE: self.thickness_initializer(
                key_thickness, self.spec.thickness_metasurface
            ),
            DENSITY_METASURFACE: self.density_initializer(
                key_density, self.seed_density
            ),
            OFFSET_MONITOR_SUBSTRATE: self.thickness_initializer(
                key_offset, self.spec.offset_monitor_substrate
            ),
        }
        # Ensure that there are no weak types in the initial parameters.
        return tree_util.tree_map(
            lambda x: jnp.asarray(x, jnp.asarray(x).dtype), params
        )

    def response(
        self,
        params: Params,
        *,
        wavelength: Optional[float | jnp.ndarray] = None,
        polar_angle: Optional[float | jnp.ndarray] = None,
        azimuthal_angle: Optional[float | jnp.ndarray] = None,
        expansion: Optional[fmmax.Expansion] = None,
        compute_fields: bool = False,
    ) -> Tuple[BayerResponse, base.AuxDict]:
        """Computes the response of the sorter.

        Args:
            params: The parameters defining the sorter, with structure matching that
                of the parameters returned by the `init` method.
            wavelength: Optional wavelength to override the default in `sim_params`.
            polar_angle: Optional polar angle to override the default.
            azimuthal_angle: Optional azimuthal angle to override the default.
            expansion: Optional expansion to override the default `expansion`.
            compute_fields: If `True`, computes fields in an xz cross section.

        Returns:
            The `(response, aux)` tuple.
        """
        if expansion is None:
            expansion = self.expansion
        if wavelength is None:
            wavelength = self.sim_params.wavelength
        if polar_angle is None:
            polar_angle = self.sim_params.polar_angle
        if azimuthal_angle is None:
            azimuthal_angle = self.sim_params.azimuthal_angle

        spec = dataclasses.replace(
            self.spec,
            thickness_metasurface=(
                params[THICKNESS_METASURFACE]  # type: ignore[arg-type]
            ),
            offset_monitor_substrate=(
                params[OFFSET_MONITOR_SUBSTRATE]  # type: ignore[arg-type]
            ),
        )
        return simulate_color_sorter(
            density=params[DENSITY_METASURFACE],  # type: ignore[arg-type]
            spec=spec,
            wavelength=jnp.asarray(wavelength),
            polar_angle=jnp.asarray(polar_angle),
            azimuthal_angle=jnp.asarray(azimuthal_angle),
            expansion=expansion,
            formulation=self.sim_params.formulation,
            compute_fields=compute_fields,
        )


def _seed_density(grid_shape: Tuple[int, int], **kwargs: Any) -> types.Density2DArray:
    """Return the seed density for a bayer sorter component.

    Args:
        grid_shape: The shape of the grid on which the density is defined.
        kwargs: keyword arguments specifying additional properties of the seed
            density, e.g. symmetries.

    Returns:
        The seed density.
    """

    # Check kwargs that are required for a sorter component.
    invalid_kwargs = ("array", "lower_bound", "upper_bound", "periodic")
    if any(k in invalid_kwargs for k in kwargs):
        raise ValueError(
            f"Attributes were specified which confict with automatically-extracted "
            f"attributes. Got {kwargs.keys()} when {invalid_kwargs} are automatically "
            f"extracted."
        )

    mid_density_value = (DENSITY_LOWER_BOUND + DENSITY_UPPER_BOUND) / 2
    return types.Density2DArray(
        array=jnp.full(grid_shape, mid_density_value),
        lower_bound=DENSITY_LOWER_BOUND,
        upper_bound=DENSITY_UPPER_BOUND,
        periodic=(True, True),
        **kwargs,
    )


def simulate_color_sorter(
    density: types.Density2DArray,
    spec: BayerSpec,
    wavelength: jnp.ndarray,
    polar_angle: jnp.ndarray,
    azimuthal_angle: jnp.ndarray,
    expansion: fmmax.Expansion,
    formulation: fmmax.Formulation,
    compute_fields: bool,
) -> Tuple[BayerResponse, base.AuxDict]:
    """Simulates a bayer color sorter.

    Args:
        density: Defines the pattern of the metasurface layer.
        spec: Defines the physical specification of the sorter.
        wavelength: The wavelength of the excitation.
        polar_angle: The polar angle of the excitation.
        azimuthal_angle: The azimuthal angle of the excitation.
        expansion: Defines the Fourier expansion for the calculation.
        formulation: Defines the FMM formulation to be used.
        compute_fields: If `True`, fields in an xz cross section are computed and
            and included in the `aux` return variable.

    Returns:
        The `BayerResponse`, and a dictionary containing the auxilliary quantities.
    """
    primitive_lattice_vectors = fmmax.LatticeVectors(
        u=spec.period_x * fmmax.X,
        v=spec.period_y * fmmax.Y,
    )
    in_plane_wavevector = fmmax.plane_wave_in_plane_wavevector(
        wavelength=wavelength,
        polar_angle=polar_angle,
        azimuthal_angle=azimuthal_angle,
        permittivity=materials.permittivity(spec.material_ambient, wavelength),
    )

    # -------------------------------------------------------------------------
    # Layer permittivities, thicknesses, and eigensolve results.
    # -------------------------------------------------------------------------

    def permittivity_fn(material_name):
        shape = (wavelength.size, 1, 1)
        return materials.permittivity(material_name, wavelength).reshape(shape)

    eigensolve_fn = functools.partial(
        fmmax.eigensolve_isotropic_media,
        wavelength=wavelength,
        in_plane_wavevector=in_plane_wavevector,
        primitive_lattice_vectors=primitive_lattice_vectors,
        expansion=expansion,
        formulation=formulation,
    )

    # Eigensolve for the uniform ambient and substrate layers.
    solve_result_ambient = eigensolve_fn(
        permittivity=permittivity_fn(spec.material_ambient)
    )
    solve_result_substrate = eigensolve_fn(
        permittivity=permittivity_fn(spec.material_substrate)
    )

    # Eigensolve for the patterned metasurface layer.
    density_array = utils.transforms.rescaled_density_array(
        density=density,
        lower_bound=DENSITY_LOWER_BOUND,
        upper_bound=DENSITY_UPPER_BOUND,
    )
    permittivity_metasurface = utils.transforms.interpolate_permittivity(
        permittivity_solid=permittivity_fn(spec.material_metasurface_solid),
        permittivity_void=permittivity_fn(spec.material_metasurface_void),
        density=density_array,
    )
    solve_result_metasurface = eigensolve_fn(permittivity=permittivity_metasurface)

    # Assemble the list of all layer solve results.
    layer_solve_results = [
        solve_result_ambient,
        solve_result_metasurface,
        solve_result_substrate,
    ]
    layer_thicknesses = [
        jnp.asarray(spec.thickness_ambient),
        jnp.asarray(spec.thickness_metasurface.array),
        jnp.asarray(spec.thickness_substrate),
    ]

    # -------------------------------------------------------------------------
    # Scattering matrix assembly.
    # -------------------------------------------------------------------------

    if compute_fields:
        # If the field calculation is desired, compute the interior scattering
        # matrices. For each layer in the stack, the interior scattering matrices
        # consist of a pair of matrices, one for the substack below the layer, and
        # one for the substack above the layer.
        s_matrices_interior = fmmax.stack_s_matrices_interior(
            layer_solve_results=layer_solve_results,
            layer_thicknesses=layer_thicknesses,
        )
        s_matrix = s_matrices_interior[-1][0]
    else:
        s_matrix = fmmax.stack_s_matrix(layer_solve_results, layer_thicknesses)

    # -------------------------------------------------------------------------
    # Excitation and wave amplitude calculation.
    # -------------------------------------------------------------------------

    n = expansion.num_terms
    assert tuple(expansion.basis_coefficients[0, :]) == (0, 0)
    assert expansion.basis_coefficients.shape[0] == n

    # Generate wave amplitudes for forward-going waves with both polarizations.
    fwd_ambient_start = jnp.zeros((2 * n, 2), dtype=complex)
    fwd_ambient_start = fwd_ambient_start.at[0, 0].set(1)
    fwd_ambient_start = fwd_ambient_start.at[n, 1].set(1)

    # Compute the backward-goingpower at the start of the ambient.
    bwd_ambient_end = s_matrix.s21 @ fwd_ambient_start
    bwd_ambient_start = fmmax.propagate_amplitude(
        bwd_ambient_end,
        jnp.asarray(spec.thickness_ambient),
        layer_solve_result=layer_solve_results[0],
    )
    sz_fwd_ambient, sz_bwd_ambient = fmmax.amplitude_poynting_flux(
        forward_amplitude=fwd_ambient_start,
        backward_amplitude=bwd_ambient_start,
        layer_solve_result=layer_solve_results[0],
    )
    sz_fwd_ambient_sum = jnp.sum(jnp.abs(sz_fwd_ambient), axis=-2)
    sz_bwd_ambient_sum = jnp.sum(jnp.abs(sz_bwd_ambient), axis=-2)
    reflection = jnp.abs(sz_bwd_ambient_sum) / jnp.abs(sz_fwd_ambient_sum)

    # Compute the forward-going and backward-going wave amplitudes in the substrate,
    # a distance `spec.offset_monitor_substrate` from the start of the substrate.
    fwd_substrate_start = s_matrix.s11 @ fwd_ambient_start
    fwd_substrate_monitor, bwd_substrate_monitor = fmmax.colocate_amplitudes(
        fwd_substrate_start,
        jnp.zeros_like(fwd_substrate_start),
        z_offset=jnp.asarray(spec.offset_monitor_substrate.array),
        layer_solve_result=layer_solve_results[-1],
        layer_thickness=layer_thicknesses[-1],
    )

    # -------------------------------------------------------------------------
    # Compute fields at the monitor plane.
    # -------------------------------------------------------------------------

    # Compute transmitted power directly from the wave amplitudes.
    sz_fwd_monitor, _ = fmmax.amplitude_poynting_flux(
        forward_amplitude=fwd_substrate_monitor,
        backward_amplitude=bwd_substrate_monitor,
        layer_solve_result=layer_solve_results[-1],
    )
    sz_fwd_monitor_sum = jnp.sum(jnp.abs(sz_fwd_monitor), axis=-2)
    transmission = jnp.abs(sz_fwd_monitor_sum) / jnp.abs(sz_fwd_ambient_sum)

    # Compute electric and magnetic fields at the monitor plane in their Fourier
    # representation, and then on the real-space grid.
    ef, hf = fmmax.fields_from_wave_amplitudes(
        forward_amplitude=fwd_substrate_monitor,
        backward_amplitude=bwd_substrate_monitor,
        layer_solve_result=layer_solve_results[-1],
    )
    grid_shape: Tuple[int, int]
    grid_shape = density.shape[-2:]  # type: ignore[assignment]
    (ex, ey, ez), (hx, hy, hz), (x, y) = fmmax.fields_on_grid(
        electric_field=ef,
        magnetic_field=hf,
        layer_solve_result=layer_solve_results[-1],
        shape=grid_shape,
        num_unit_cells=(1, 1),
    )
    batch_shape = layer_solve_results[0].batch_shape
    assert ex.shape == batch_shape + grid_shape + (2,)

    # Compute the Poynting flux on the real-space grid at the monitor plane.
    sz = fmmax.time_average_z_poynting_flux((ex, ey, ez), (hx, hy, hz))
    assert sz.shape == batch_shape + grid_shape + (2,)

    # Create masks selecting the four pixelss.
    pixel_mask = _pixel_mask(grid_shape=grid_shape)
    assert pixel_mask.shape == grid_shape + (1, 4)

    # Use the mask to compute the time average Poynting flux into each quadrant. The
    # trailing two dimensions have shape `(2, 4)`; index `(i, j)` corresponds
    # to power for the `i` excitation (i.e. polarization) in the `j` pixel.
    pixel_transmission = jnp.mean(pixel_mask * sz[..., jnp.newaxis], axis=(-4, -3))
    pixel_transmission /= sz_fwd_ambient_sum[..., jnp.newaxis]

    assert pixel_transmission.shape == batch_shape + (2, 4)

    response = BayerResponse(
        wavelength=wavelength,
        polar_angle=polar_angle,
        azimuthal_angle=azimuthal_angle,
        transmission=pixel_transmission,
        reflection=reflection,
    )

    aux = {
        EFIELD_XY: (ex, ey, ez),
        HFIELD_XY: (hx, hy, hz),
        POYNTING_FLUX_XY: sz,
        COORDINATES_XY: (x, y),
        TRANSMITTED_POWER: transmission,
    }

    # -------------------------------------------------------------------------
    # Optionally compute fields in an xz cross section.
    # -------------------------------------------------------------------------

    if compute_fields:
        amplitudes_interior = fmmax.stack_amplitudes_interior(
            s_matrices_interior=s_matrices_interior,
            forward_amplitude_0_start=fwd_ambient_start,
            backward_amplitude_N_end=jnp.zeros_like(fwd_ambient_start),
        )
        layer_znum = tuple(
            [int(jnp.round(t / spec.grid_spacing) + 1) for t in layer_thicknesses]
        )
        x = jnp.stack(
            [jnp.arange(0, spec.grid_shape[0]) * spec.grid_spacing] * 2,
            axis=-1,
        )
        y = jnp.stack(
            [
                jnp.ones((spec.grid_shape[0],)) * spec.pixel_size * 0.5,
                jnp.ones((spec.grid_shape[0],)) * spec.pixel_size * 1.5,
            ],
            axis=-1,
        )
        assert x.shape == y.shape == (spec.grid_shape[0], 2)

        (
            (ex, ey, ez),
            (hx, hy, hz),
            (xf, yf, zf),
        ) = fmmax.stack_fields_3d_on_coordinates(
            amplitudes_interior=amplitudes_interior,
            layer_solve_results=layer_solve_results,
            layer_thicknesses=layer_thicknesses,
            layer_znum=layer_znum,
            x=x,
            y=y,
        )
        aux.update(
            {
                EFIELD_XZ: (ex, ey, ez),
                HFIELD_XZ: (hx, hy, hz),
                COORDINATES_XZ: (xf, yf, zf),
            }
        )

    return response, aux


# -----------------------------------------------------------------------------
# Functions related to the densities of various layers in the stack.
# -----------------------------------------------------------------------------


def _pixel_mask(grid_shape: Tuple[int, int]) -> jnp.ndarray:
    """Return masks that select the four quadrants of a sorter.

    The quadrants are numbered as follows:     0  |  1
                                             -----------
                                               2  |  3

    Args:
        grid_shape: The shape of the grid for which to return the mask.

    Returns:
        The quadrant mask, with shape `grid_shape + (1, 4)`.
    """
    xdim = grid_shape[0] // 2
    ydim = grid_shape[1] // 2
    ones = jnp.ones((xdim, ydim))
    zeros = jnp.zeros((xdim, ydim))
    quadrant_mask = jnp.stack(
        [
            jnp.block([[ones, zeros], [zeros, zeros]]),
            jnp.block([[zeros, ones], [zeros, zeros]]),
            jnp.block([[zeros, zeros], [ones, zeros]]),
            jnp.block([[zeros, zeros], [zeros, ones]]),
        ],
        axis=-1,
    )
    quadrant_mask = quadrant_mask[:, :, jnp.newaxis, :]
    assert quadrant_mask.shape == grid_shape + (1, 4)
    return quadrant_mask



"""Base types for challenges and components.

Copyright (c) 2025 invrs.io LLC
"""

import abc
from typing import Any, Callable, Dict, Tuple

import fmmax
import jax
import jax.numpy as jnp
from totypes import json_utils, types

from invrs_gym.utils import metrics

AuxDict = Dict[str, Any]
PyTree = Any
DensityInitializer = Callable[[jax.Array, types.Density2DArray], types.Density2DArray]
ThicknessInitializer = Callable[[jax.Array, types.BoundedArray], types.BoundedArray]

BINARIZATION_DEGREE = "binarization_degree"


class Component(abc.ABC):
    """Base class for components.

    The component represents the physical structure to be optimized. Initial parameters
    are returned by the `init` method; these can include types having metadata that
    specifies bounds on parameter values or target characteristics, e.g. a minimum
    feature size.

    The component has a default "excitation", e.g. illumination with a particular
    wavelength of light from a specific direction; the response of the component to the
    excitation is obtained with the `response` method. The `response` method of classes
    inheriting from `Component` will generally have optional arguments to modify the
    excitation condition.
    """

    @abc.abstractmethod
    def init(self, key: jax.Array):
        """Returns the initial parameters given a `PRNGKey`."""

    @abc.abstractmethod
    def response(self, params: PyTree) -> Tuple[Any, AuxDict]:
        """Returns the response of the component for the given `params`."""


class Challenge(abc.ABC):
    """Base class for challenges.

    Challenges consist of a `Component` to be optimized and a loss function which
    returns a scalar loss given the component response.

    Challenges also include an `eval_metric` function whcih computes a scalar metric
    (independent of the loss), which can be used to evaluate the quality of solutions
    to the challenge. Higher `eval_metric` corresponds to a better solution.

    Challenges also include a `metrics` function which computes additional quantities
    that may be useful assessing the quality of a particular component.

    Attributes:
        component: The `Component` to be optimized.
    """

    component: Component

    @abc.abstractmethod
    def loss(self, response: Any) -> jnp.ndarray:
        """Compute scalar loss for the `response`."""

    @abc.abstractmethod
    def eval_metric(self, response: Any) -> jnp.ndarray:
        """Compute scalar evaluation metric for the `response`."""

    def metrics(self, response: Any, params: PyTree, aux: AuxDict) -> AuxDict:
        """Compute metrics for a component response and associated quantities."""
        del response, aux
        return {BINARIZATION_DEGREE: metrics.binarization_degree(params)}


# Several challenges use the `fmmax` simulator, and contain `fmmax` custom objects in
# their responses. Ensure these are serializable by registering with totypes.
json_utils.register_custom_type(fmmax.Expansion)
json_utils.register_custom_type(fmmax.LatticeVectors)



"""Defines the bayer color sorter challenge.

Copyright (c) 2025 invrs.io LLC
"""

import dataclasses
import functools
from typing import Sequence, Tuple

import fmmax
from jax import nn
from jax import numpy as jnp
from totypes import symmetry, types

from invrs_gym import utils
from invrs_gym.challenges import base
from invrs_gym.challenges.bayer import component as bayer_component
from invrs_gym.utils import materials

EFFICIENCY_BLUE_MEAN = "efficiency_blue_mean"
EFFICIENCY_GREEN_MEAN = "efficiency_green_mean"
EFFICIENCY_RED_MEAN = "efficiency_red_mean"
CROSSTALK_BLUE_MEAN = "crosstalk_blue_mean"
CROSSTALK_GREEN_MEAN = "crosstalk_green_mean"
CROSSTALK_RED_MEAN = "crosstalk_red_mean"

# Cutoff between UV/blue wavelengths, blue/green wavelengths, etc. in microns.
UV_BLUE_CUTOFF = 0.4
BLUE_GREEN_CUTOFF = 0.5
GREEN_RED_CUTOFF = 0.6
RED_IR_CUTOFF = 0.7


density_initializer = functools.partial(
    utils.initializers.noisy_density_initializer,
    relative_mean=0.5,
    relative_noise_amplitude=0.1,
)


@dataclasses.dataclass
class BayerChallenge(base.Challenge):
    """Defines the bayer color sorter challenge."""

    component: bayer_component.BayerComponent

    def loss(
        self,
        response: bayer_component.BayerResponse,
        efficiency_target: float = 0.55,
    ) -> jnp.ndarray:
        """Compute a scalar loss from the component `response`.

        Args:
            response: The response obtained from simulating the component.
            efficiency_target: Value used in a nonlinearity within the loss function.
                When the efficiency value is far from the target, the dependence of
                loss on efficiency is larger.

        Returns:
            The scalar loss value.
        """
        # Average over the two polarizations.
        transmission = jnp.mean(response.transmission, axis=-2)
        assert transmission.shape == (response.wavelength.size, 4)

        is_blue, is_green, is_red = _detect_color(response.wavelength)
        t_blue, t_green, t_red = _transmission_per_pixel(transmission)

        efficiency_blue = jnp.sum(t_blue * is_blue) / jnp.sum(is_blue)
        efficiency_green = jnp.sum(t_green * is_green) / jnp.sum(is_green)
        efficiency_red = jnp.sum(t_red * is_red) / jnp.sum(is_red)

        denom = 1 - efficiency_target
        loss_blue = nn.softplus((efficiency_target - efficiency_blue) / denom)
        loss_green = nn.softplus((efficiency_target - efficiency_green) / denom)
        loss_red = nn.softplus((efficiency_target - efficiency_red) / denom)
        return loss_blue + loss_green + loss_red

    def eval_metric(self, response: bayer_component.BayerResponse) -> jnp.ndarray:
        """Compute eval metric from the component `response`.

        The eval metric for the bayer sorter is the minimum of the red, green, and
        blue efficiencies, where the efficiency fo a given color is the average
        efficiency with which wavelengths associated with the color are coupled into
        the appropriate subpixel(s).

        Args:
            response: The component response.

        Returns:
            The scalar eval metric.
        """
        # Average over the two polarizations.
        transmission = jnp.mean(response.transmission, axis=-2)
        assert transmission.shape == (response.wavelength.size, 4)

        is_blue, is_green, is_red = _detect_color(response.wavelength)
        t_blue, t_green, t_red = _transmission_per_pixel(transmission)
        efficiency_blue = jnp.sum(t_blue * is_blue) / jnp.sum(is_blue)
        efficiency_green = jnp.sum(t_green * is_green) / jnp.sum(is_green)
        efficiency_red = jnp.sum(t_red * is_red) / jnp.sum(is_red)
        return jnp.amin(
            jnp.asarray([efficiency_blue, efficiency_green, efficiency_red])
        )

    def metrics(
        self,
        response: bayer_component.BayerResponse,
        params: types.Density2DArray,
        aux: base.AuxDict,
    ) -> base.AuxDict:
        """Compute challenge metrics."""
        metrics = super().metrics(response, params, aux)

        # Average over the two polarizations.
        transmission = jnp.mean(response.transmission, axis=-2)
        assert transmission.shape == (response.wavelength.size, 4)

        is_blue, is_green, is_red = _detect_color(response.wavelength)
        t_blue, t_green, t_red = _transmission_per_pixel(transmission)
        efficiency_blue = jnp.sum(t_blue * is_blue) / jnp.sum(is_blue)
        efficiency_green = jnp.sum(t_green * is_green) / jnp.sum(is_green)
        efficiency_red = jnp.sum(t_red * is_red) / jnp.sum(is_red)

        # Crosstalk for a given color is total unwanted power in that color's pixel.
        crosstalk_blue = jnp.sum(t_blue * ~is_blue) / jnp.sum(~is_blue)
        crosstalk_green = jnp.sum(t_green * ~is_green) / jnp.sum(~is_green)
        crosstalk_red = jnp.sum(t_red * ~is_red) / jnp.sum(~is_red)

        transmitted_power = aux[bayer_component.TRANSMITTED_POWER]
        assert response.reflection.shape == transmitted_power.shape

        transmitted_power = jnp.sum(response.transmission, axis=-1)
        power = transmitted_power + response.reflection
        assert power.shape == response.wavelength.shape + (2,)

        metrics.update(
            {
                EFFICIENCY_BLUE_MEAN: efficiency_blue,
                EFFICIENCY_GREEN_MEAN: efficiency_green,
                EFFICIENCY_RED_MEAN: efficiency_red,
                CROSSTALK_BLUE_MEAN: crosstalk_blue,
                CROSSTALK_GREEN_MEAN: crosstalk_green,
                CROSSTALK_RED_MEAN: crosstalk_red,
            }
        )
        return metrics


def _detect_color(
    wavelength: jnp.ndarray,
) -> Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray]:
    """Detects color (blue, green, red) from `wavelength`."""
    is_blue = (wavelength >= UV_BLUE_CUTOFF) & (wavelength < BLUE_GREEN_CUTOFF)
    is_green = (wavelength >= BLUE_GREEN_CUTOFF) & (wavelength < GREEN_RED_CUTOFF)
    is_red = (wavelength >= GREEN_RED_CUTOFF) & (wavelength < RED_IR_CUTOFF)
    return is_blue, is_green, is_red


def _transmission_per_pixel(
    transmission: jnp.ndarray,
) -> Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray]:
    """Computes transmission into the blue, green, and red pixels."""
    assert transmission.shape[-1] == 4
    blue_transmission = transmission[..., 0]
    green_transmission = transmission[..., 1] + transmission[..., 2]
    red_transmission = transmission[..., 3]
    return blue_transmission, green_transmission, red_transmission


SYMMETRIES = (symmetry.REFLECTION_NW_SE,)

# Minimum width and spacing are 80 nm for the default dimensions.
MINIMUM_WIDTH = 8
MINIMUM_SPACING = 8


BAYER_SPEC = bayer_component.BayerSpec(
    material_ambient=materials.VACUUM,
    material_metasurface_solid=materials.SI3N4,
    material_metasurface_void=materials.VACUUM,
    material_substrate=materials.SIO2,
    thickness_ambient=1.0,
    thickness_metasurface=types.BoundedArray(0.6, lower_bound=0.4, upper_bound=0.8),
    thickness_substrate=3.0,
    pixel_size=1.0,
    grid_spacing=0.01,
    offset_monitor_substrate=types.BoundedArray(2.4, lower_bound=2.0, upper_bound=2.5),
)

BAYER_SIM_PARAMS = bayer_component.BayerSimParams(
    wavelength=jnp.array([0.45, 0.55, 0.65]),
    polar_angle=0.0,
    azimuthal_angle=0.0,
    formulation=fmmax.Formulation.JONES_DIRECT_FOURIER,
    approximate_num_terms=600,
    truncation=fmmax.Truncation.CIRCULAR,
)


def bayer_sorter(
    minimum_width: int = MINIMUM_WIDTH,
    minimum_spacing: int = MINIMUM_SPACING,
    thickness_initializer: bayer_component.ThicknessInitializer = (
        utils.initializers.identity_initializer
    ),
    density_initializer: base.DensityInitializer = density_initializer,
    spec: bayer_component.BayerSpec = BAYER_SPEC,
    sim_params: bayer_component.BayerSimParams = BAYER_SIM_PARAMS,
    symmetries: Sequence[str] = SYMMETRIES,
) -> BayerChallenge:
    """The bayer color sorter challenge.

    The bayer sorter challenge entails the design of a color-sorting metasurface that
    replaces the color filter array in a conventional image sensor. The thickness of
    the metasurface, and the distance between metasurface and focal plane are further
    degrees of freedom. The bayer challenge is based on "Pixel-level Bayer-type colour
    router based on metasurfaces" by Zou et al.

    https://www.nature.com/articles/s41467-022-31019-7

    Args:
        minimum_width: The minimum width target for the challenge, in pixels.  The
            default value of 8 corresponds to a physical size of approximately 80 nm.
        minimum_spacing: The minimum spacing target for the challenge, in pixels.
        thickness_initializer: Callable which returns the initial thickness, given a
            key and seed thickness.
        density_initializer: Callble which returns the initial density, given a
            key and seed density.
        spec: Defines the physical specification of the bayer sorter.
        sim_params: Defines the simulation settings of the bayer sorter.
        symmetries: Defines the symmetries of the bayer sorter.

    Returns:
        The `BayerChallenge`.
    """
    return BayerChallenge(
        component=bayer_component.BayerComponent(
            spec=spec,
            sim_params=sim_params,
            thickness_initializer=thickness_initializer,
            density_initializer=density_initializer,
            minimum_width=minimum_width,
            minimum_spacing=minimum_spacing,
            symmetries=symmetries,
        ),
    )



"""Defines functions that compute material properties.

Copyright (c) 2025 invrs.io LLC
"""

import functools
import pathlib
import warnings
from typing import Dict, Protocol, Union

import jax
import numpy as onp
import refractiveindex2 as ri
from jax import numpy as jnp
from packaging import version

Material = str | ri.RefractiveIndexMaterial

# Define some common materials.
SI = ri.RefractiveIndexMaterial(shelf="main", book="Si", page="Green-1995")
AMORPHOUS_SI = ri.RefractiveIndexMaterial(shelf="main", book="Si", page="Pierce")
SIO2 = ri.RefractiveIndexMaterial(shelf="main", book="SiO2", page="Malitson")
SI3N4 = ri.RefractiveIndexMaterial(shelf="main", book="Si3N4", page="Luke")
TIO2 = ri.RefractiveIndexMaterial(shelf="main", book="TiO2", page="Jolivet-amorphous")
VACUUM = "vacuum"  # Permittivity is computed via dedicated function.


if version.Version(jax.__version__) > version.Version("0.4.31"):
    callback = functools.partial(jax.pure_callback, vmap_method="broadcast_all")
else:
    callback = functools.partial(jax.pure_callback, vectorized=True)


def permittivity(
    material: str | ri.RefractiveIndexMaterial,
    wavelength_um: float | jnp.ndarray,
    background_extinction_coeff: float = 0.0,
) -> jnp.ndarray:
    """Return the permittivity for the specified material.

    Args:
        material: The name of a registered material, or a `ri.RefractiveIndexMaterial`
            for which the permittivity is sought.
        wavelength_um: The wavelength for the permittivity, in units of microns.
        background_extinction_coeff: Additional extinction coefficient to be included
            in the permittivity. Positive values correspond to optical loss, and the
            default value is 0.

    Returns:
        The permittivity of the material at the specified wavelengths.
    """
    wavelength_um = jnp.asarray(wavelength_um)
    if isinstance(material, ri.RefractiveIndexMaterial):
        permittivity_fn: PermittivityFn = functools.partial(
            permittivity_from_database,
            material=material,
        )
    else:
        permittivity_fn = PERMITTIVITY_FNS[material]
    return permittivity_fn(
        wavelength_um=wavelength_um,
        background_extinction_coeff=background_extinction_coeff,
    )


def permittivity_from_database(
    material: ri.RefractiveIndexMaterial,
    wavelength_um: float | jnp.ndarray,
    background_extinction_coeff: float,
) -> jnp.ndarray:
    """Return the permittivity for the specified material from the database."""
    wavelength_um = jnp.asarray(wavelength_um)

    def _refractive_index_fn(wavelength_um: jnp.ndarray) -> onp.ndarray:
        numpy_wavelength_um = onp.asarray(wavelength_um)
        dtype = onp.promote_types(wavelength_um.dtype, onp.complex64)
        try:
            epsilon = material.get_epsilon(numpy_wavelength_um)
            refractive_index = onp.sqrt(epsilon)
        except ri.refractiveindex.NoExtinctionCoefficient:
            refractive_index = material.get_refractive_index(numpy_wavelength_um)
        return onp.asarray(refractive_index, dtype=dtype)

    dtype = jnp.promote_types(wavelength_um.dtype, jnp.complex64)
    result_shape_dtypes = jnp.zeros_like(wavelength_um, dtype=dtype)
    refractive_index = callback(
        _refractive_index_fn, result_shape_dtypes, wavelength_um
    )
    return (refractive_index + 1j * background_extinction_coeff) ** 2


def permittivity_vacuum(
    wavelength_um: float | jnp.ndarray,
    background_extinction_coeff: float = 0.0,
) -> jnp.ndarray:
    """Return the permittivity of vacuum, with optional background extinction coeff."""
    wavelength_um = jnp.asarray(wavelength_um)
    dtype = jnp.promote_types(wavelength_um.dtype, jnp.complex64)
    return jnp.full(
        wavelength_um.shape, 1.0 + 1j * background_extinction_coeff, dtype=dtype
    )


class PermittivityFn(Protocol):
    def __call__(
        self, wavelength_um: jnp.ndarray, background_extinction_coeff: float
    ) -> jnp.ndarray:
        ...


PERMITTIVITY_FNS: Dict[str, PermittivityFn] = {
    VACUUM: permittivity_vacuum,
}


def register_material(name: str, path: Union[str, pathlib.Path]) -> None:
    """Registers a material so it is accessible by `get_permittivity` function."""
    with jax.ensure_compile_time_eval():
        data_wvl_n_k = onp.genfromtxt(path, delimiter=",", comments="#")
        data_wavelength_um = jnp.asarray(data_wvl_n_k[:, 0])
        data_permittivity = (data_wvl_n_k[:, 1] + 1j * data_wvl_n_k[:, 2]) ** 2

        if name in PERMITTIVITY_FNS:
            warnings.warn(f"Material {name} already registered.")

        def _permittivity_fn(
            wavelength_um: float | jnp.ndarray,
            background_extinction_coeff: float,
        ) -> jnp.ndarray:
            wavelength_um = jnp.asarray(wavelength_um)
            dtype = jnp.promote_types(wavelength_um.dtype, jnp.complex64)
            refractive_index = jnp.sqrt(
                jnp.interp(wavelength_um, data_wavelength_um, data_permittivity)
            )
            return jnp.asarray(
                (refractive_index + 1j * background_extinction_coeff) ** 2, dtype=dtype
            )

        PERMITTIVITY_FNS[name] = _permittivity_fn
